# Generated by roxygen2: do not edit by hand

export(AT2TS)
export(DT2TS)
export(IML2IMW)
export(PSL2PSW)
export(PSW2PSL)
export(TS2IMF)
export(TSL2AI)
export(TSL2CAV)
export(TSL2CAV5)
export(TSL2IM)
export(TSL2PS)
export(TSL2TSW)
export(TSW2TSL)
export(VT2TS)
export(alignComponents)
export(archiveRawOwner)
export(auditDistances)
export(auditParsers)
export(auditSite)
export(buildRawFileTable)
export(buildRawIntensityTable)
export(buildRawRecordTable)
export(extractRecord)
export(fitModalFactor)
export(fitRecordFactor)
export(getIntensity)
export(getRawIntensities)
export(mapComponents)
export(matchCompatibility)
export(normalizeTS)
export(parseRecord)
export(readAC)
export(readAT)
export(readAT2)
export(readDT)
export(readISEE)
export(readTR)
export(readTS)
export(readTwoCol)
export(readV2)
export(readV2A)
export(readVT)
export(runExport)
export(runGMSP)
export(runIMF)
export(runMatch)
export(runPSA)
export(runPlot)
export(runProcess)
export(runSelect)
export(runStage0)
export(runTrim)
export(selectRecords)
export(writeSelection)
import(data.table)
importFrom(EMD,emd)
importFrom(VMDecomp,vmd)
importFrom(data.table,":=")
importFrom(data.table,as.data.table)
importFrom(data.table,copy)
importFrom(data.table,data.table)
importFrom(data.table,dcast)
importFrom(data.table,fcase)
importFrom(data.table,fifelse)
importFrom(data.table,fread)
importFrom(data.table,fwrite)
importFrom(data.table,is.data.table)
importFrom(data.table,melt)
importFrom(data.table,rbindlist)
importFrom(data.table,setattr)
importFrom(data.table,setcolorder)
importFrom(data.table,setnames)
importFrom(data.table,setorder)
importFrom(data.table,setorderv)
importFrom(digest,digest)
importFrom(hht,CEEMD)
importFrom(hht,EEMD)
importFrom(hht,EEMDCompile)
importFrom(jsonlite,fromJSON)
importFrom(jsonlite,read_json)
importFrom(jsonlite,write_json)
importFrom(openssl,md5)
importFrom(pracma,detrend)
importFrom(purrr,map)
importFrom(seewave,istft)
importFrom(seewave,stdft)
importFrom(signal,resample)
importFrom(spectral,spec.fft)
importFrom(stats,cov)
importFrom(stats,fft)
importFrom(stats,median)
importFrom(stats,na.omit)
importFrom(stats,optim)
importFrom(stats,setNames)
importFrom(stringr,str_split)
importFrom(utils,tail)
importFrom(utils,tar)
importFrom(utils,untar)
