# scripts/gmsp/global.R -- shared GMSP pipeline values.

PATH.records <- "gmsp/raw"

KEY.record <- c("RecordID", "OwnerID", "EventID", "StationID")
FILES.TS <- c("TSW.csv", "IMW.csv", "OCIDMAP.csv", "META.csv")
FILES.IMF <- c("IMF.csv", "IMFF.csv")
